from typing import List, Literal, Optional, Union

from pydantic import BaseModel, Field

from bananalyzer.data.schemas import GoalType

XDistDistributionMode = Literal[
    "load", "loadscope", "loadfile", "loadgroup", "worksteal", "no"
]


class PytestArgs(BaseModel):
    s: bool
    q: bool
    xml: Optional[str] = Field(description="Path to the xml report file")


class XDistArgs(BaseModel):
    dist: XDistDistributionMode = Field(description="Distribution mode (xdist)")
    n: Union[int, Literal["logical", "auto"]] = Field(
        description="Number of workers (xdist)"
    )


class Args(BaseModel):
    path: Union[str, Literal["DOWNLOAD_ONLY"]]
    headless: bool
    single_browser_instance: bool
    id: Optional[List[str]] = Field(default=None)
    tags: Optional[List[str]] = Field(default=None)
    domain: Optional[str] = Field(default=None)
    intent: Optional[GoalType] = Field(default=None)
    category: Optional[str] = Field(default=None)
    subcategory: Optional[str] = Field(default=None)
    skip: List[str]
    type: Optional[str] = Field(default=None)
    download: bool
    examples_bucket: Optional[str] = Field(default=None)
    test: bool
    count: Optional[int]
    pytest_args: PytestArgs
    xdist_args: XDistArgs


class AgentRunnerClass(BaseModel):
    class_name: str
    class_path: str


MARKER_PREFIX = "bananalyzer_"



import json
from pathlib import Path
from typing import Any, Dict

from nanoid import generate


def load_json_data(file_path: str) -> Dict[str, Any]:
    with open(file_path, 'r') as file:
        return json.load(file)


def save_json_data(file_path: str, data: Dict[str, Any]) -> None:
    with open(file_path, 'w') as file:
        json.dump(data, file, indent=4)


def create_nano_id() -> str:
    disallowed = ['-', '_']
    nano_id = generate()
    while any([item in nano_id for item in disallowed]):
        nano_id = generate()
    return nano_id


def replace_uuids_with_nanoids(json_file_path: str, directory_path: str) -> None:
    data = load_json_data(json_file_path)
    for entry in data:
        old_id = entry['id']
        new_id = create_nano_id()
        entry['id'] = new_id

        # Rename the folder
        old_folder = Path(directory_path) / old_id
        new_folder = Path(directory_path) / new_id

        if old_folder.exists():
            old_folder.rename(new_folder)

    save_json_data(json_file_path, data)

# Usage
json_file_path = '/Users/asimshrestha/Documents/GitHub/bananalyzer/static/examples.json'  # Path to your JSON file
directory_path = '/Users/asimshrestha/Documents/GitHub/bananalyzer/static'  # Path to the directories
replace_uuids_with_nanoids(json_file_path, directory_path)



import json
from typing import Any, Dict, List


def download_examples_from_s3(examples_bucket: str) -> List[Dict[str, Any]]:
    import boto3

    s3 = boto3.client("s3", region_name="us-east-1")

    examples = []
    response = s3.list_objects(Bucket=examples_bucket)
    for file in response["Contents"]:
        key = file["Key"]

        if key.startswith("example-") and key.endswith(".json"):
            response = s3.get_object(Bucket=examples_bucket, Key=key)
            file_content = response["Body"].read().decode("utf-8")
            example = json.loads(file_content)

            if example["fetch_id"] == "":
                del example["fetch_id"]

            for row in example["evals"][0]["expected"]:
                row = {k: v for k, v in row.items() if not k.startswith("__")}
                if "context" in row:
                    for key, value in row["context"].items():
                        row[key] = value
                    del row["context"]

            examples.append(example)

    return examples


def download_mhtml(url: str) -> str:
    import boto3

    s3 = boto3.client("s3", region_name="us-east-1")

    if url.startswith("s3://"):
        bucket_name = url.split("/")[2]
        key = "/".join(url.split("/")[3:])

        response = s3.get_object(Bucket=bucket_name, Key=key)
        mhtml = response["Body"].read().decode("utf-8")

        return mhtml
    else:
        raise NotImplementedError("Only s3:// URIs are currently supported")



from typing import Dict, Type, Union, Any

from pydantic import BaseModel, Field

from bananalyzer.data.schemas import FetchId

"""
This file contains mapping of fetch_id to fetch schema to avoid duplicate schemas in examples.json
"""

CONTACT_SCHEMA_GOAL = "Fetch the main hospital location contact information on the current page. There may be multiple location information or sub location information on the page. Only fetch the main location information which should be at the top of the page."


class ContactSchema(BaseModel):
    name: str = Field(
        description="Name of the location, facility, or the service provided by the clinic. Typically available at the top of the page or in the contact section. Do NOT use the address as the name which may be street/ state like 'Henry Adams, SF'. Double check that this is not the case"
    )
    address: str = Field(
        description="The COMPLETE address of the location using inner text of address elements. This value MUST be the combination of the building name or medical center name, street, city, state, and ZIP. EXAMPLE: `John Ivy Medical Center\n199 Test Street\nTest, CA 94103.` Concatenate multiple elements together as needed and retain formatting if possible. The building name may be placed above the other address elements. You MUST include the building name if it is available and presented this way. Do not forget ANY of the address elements. Filter out extra words at the begining like 'Address:' or extra words at the end like 'Phone', 'Fax', 'Directions', etc. Do not use value from additional locations.",
    )
    phone: str = Field(
        description="The primary phone number of the primary location. Ensure it is the phone number of the main location. This should be positioned higher than other phone numbers. Only include the number but retain its formatting. Strip all leading or traling words like 'Phone'. Do not phone value from related locations. Ensure the used phone value is positioned close to other contact fields.",
    )
    fax: str = Field(
        description="The primary FAX number of the location. Only include the number but retain its formatting by stripping all leading or traling words like 'Fax'. Ensure the fax number you select is LABELED as a FAX number on the page via text or an icon. If there is no number labeled as fax on the page, this value MUST be left as NULL. Never assume unlabeled numbers are the fax number. Do not use value from related locations.",
    )


GOVERNMENT_CONTRACT_GOAL = "Fetch the main government contract/notice/solicitation information on the current website."


class File(BaseModel):
    title: str
    url: str


class GovernmentContractSchema(BaseModel):
    id: str = Field(description="Unique identifier for the contract")
    title: str = Field(description="Title or name of the contract")
    description: str = Field(
        description="Description or synopsis field. Combine the solicitation summary and additional instructions section / process.",
    )
    location: str = Field(
        description="Location of the issuer. May be a combination of city and state",
    )
    type: str = Field(
        description="Type of contract. May be placed under `Solicitation Type`, `Opportunity Type`, `Market Type`, etc. Not a 'status' field",
    )
    category: str = Field(description="Category the contract falls under if given")

    posted_date: str = Field(
        description="Date the contract was made available for bidding. NOT the 'effective', 'start', or 'award' date.",
    )
    due_date: str = Field(
        description="Date the contract closes for bidding. NOT the end term date.",
    )

    buyer_name: str = Field(
        description="Name of the company, organization, or agency that issued the contract. NOT a person's name."
    )
    buyer_contact_name: str = Field(
        description="Name of the specific individual that is leading the contract, if available"
    )
    buyer_contact_number: str = Field(description="Contact number of the issuer")
    buyer_contact_email: str = Field(description="Contact email of the issuer")

    attachments: list[File] = Field(
        default_factory=list,
        description="A list of all of the files/documents attached to the contract (e.g. hyperlinks to PDF's)",
    )


JobPostingSchema = {
    "job_id": {
        "type": "string",
        "description": "Unique alphanumeric identifier for the job posting.",
    },
    "company_name": {
        "type": "string",
        "description": "Name of the company offering the job.",
    },
    "company_description": {
        "type": "string",
        "description": "A brief description of the company within the job post.",
    },
    "department": {
        "type": "string",
        "description": "The department or team name within the company for the job position.",
    },
    "job_title": {
        "type": "string",
        "description": "The entire job title, including team name or specialization if present in the title.",
    },
    "job_description": {
        "type": "string",
        "description": "Comprehensive job description including all its sentences.",
    },
    "location": {"type": "string", "description": "Entire location of the job."},
    "salary_range": {
        "type": "object",
        "properties": {
            "min": {"type": "string", "description": "Minimum salary offered."},
            "max": {"type": "string", "description": "Maximum salary offered."},
            "currency": {
                "type": "string",
                "description": "The currency of the salary.",
            },
        },
    },
    "date_posted": {
        "type": "string",
        "description": "Date posted, only if present on the page.",
    },
    "apply_url": {
        "type": "string",
        "description": "The URL where applicants can apply for the job.",
    },
    "job_benefits": {
        "type": "string",
        "description": "A list of benefits provided with the job.",
    },
    "qualifications": {
        "type": "string",
        "description": "A list of required qualifications or certifications for the job.",
    },
    "preferred_qualifications": {
        "type": "string",
        "description": "A list of preferred (but not mandatory) qualifications for the job.",
    },
    "role": {
        "type": "string",
        "description": "Details about the role including responsibilities and required skills.",
    },
    "skills": {
        "type": "string",
        "description": "A list of knowledge, skills or abilities required for the job.",
    },
    "recruiter_email": {
        "type": "string",
        "description": "Email address of the recruiter or hiring manager for contact.",
    },
    "application_deadline": {
        "type": "string",
        "description": "The deadline for submitting job applications.",
    },
    "employment_type": {
        "type": "string",
        "description": "The type of employment (e.g., full-time, part-time, contract).",
    },
}


class Specification(BaseModel):
    label: str
    value: str


class Document(BaseModel):
    url: str
    filename: str


class ManufacturingCommerceSchema(BaseModel):
    mpn: str
    alias_mpns: list[str] = Field(description="Other MPNs that this part is known by")
    manufacturer: str
    classifications: list[str]
    description: str
    hero_image: str
    series: str
    lifecycle_status: str
    country_of_origin: str
    aecq_status: str
    reach_status: str
    rohs_status: str
    export_control_class_number: str
    packaging: str
    power_rating: str
    voltage_rating: str
    mount_type: str
    moisture_sensitivity_level: str
    tolerance: str
    inductance: str
    capacitance: str
    resistance: str
    min_operating_temperature: str
    max_operating_temperature: str
    leadfree: str
    termination_type: str
    num_terminations: int
    specs: list[Specification]
    product_change_notification_documents: list[Document]
    reach_compliance_documents: list[Document]
    rohs_compliance_documents: list[Document]
    datasheets: list[Document]
    specsheets: list[Document]
    suggested_alternative_mpns: list[str]


class ForumSchema(BaseModel):
    author: str = Field(description="Author of the main or original post")
    title: str = Field(description="Title of the original post")
    post_date: str
    content: str = Field(
        description="Entire content of the original post, including all sentences"
    )
    up_votes: int = Field(
        description="Number of likes, upvotes, etc on the original post"
    )
    down_votes: int
    views: int
    num_comments: int = Field(
        description="Number of comments or responses, not including the original post"
    )


class AttorneyExperience(BaseModel):
    title: str = Field(description="Title at the previous law firm")
    firm: str = Field(description="Name of the law firm")
    dates: str = Field(description="Duration of employment at the law firm")


class AttorneyBarAdmission(BaseModel):
    state: str = Field(description="State of bar admission")
    year: str = Field(description="Year of bar admission")
    country: str = Field(description="Country of bar admission", default="USA")


class AttorneyEducation(BaseModel):
    school: str = Field(description="Name of the educational institution")
    year: str = Field(description="Year of graduation")
    degree: str = Field(
        description="Type of degree obtained",
    )
    honors: str = Field(
        description="Honors received during education",
    )


class AttorneyAward(BaseModel):
    date: str = Field(description="Date or duration when the award was received")
    award: str = Field(description="Name of the award")
    url: str = Field(
        description="URL link to the award recognition",
    )


class AttorneySchema(BaseModel):
    website: str = Field(description="URL of the attorney's profile page")
    name: str = Field(description="Full name of the attorney")
    title: str = Field(
        description="Title of the attorney, such as Associate, Counsel, or Partner"
    )
    practice_areas_main: list[str] = Field(
        description="Primary practice areas of the attorney"
    )
    practice_areas_all: list[str] = Field(description="Complete list of practice areas")
    specialties: list[str] = Field(description="Specialized industry sectors")
    email: str = Field(description="Email address of the attorney")
    location: str = Field(description="Office location of the attorney")
    phone: str = Field(
        description="Direct phone number of the attorney",
    )
    bio: str = Field(description="Main bio description of the attorney")
    experience: list[AttorneyExperience] = Field(
        description="Past work history at other law firms"
    )
    matters: str = Field(description="List of past cases and work done by the attorney")
    bar_admissions: list[AttorneyBarAdmission] = Field(
        description="Bar admissions of the attorney"
    )
    law_school: list[AttorneyEducation] = Field(
        description="Law school information of the attorney"
    )
    other_schools: list[AttorneyEducation] = Field(
        description="Other education details of the attorney"
    )
    awards: list[AttorneyAward] = Field(
        description="Awards and recognitions received by the attorney"
    )
    pdf_url: str = Field(
        description="Link to a PDF bio of the attorney",
    )
    photo_url: str = Field(description="Link to the photo of the attorney")
    news: list[str] = Field(
        description="Links to news articles involving the attorney", default=[]
    )


class AttorneyJobPostingSchema(BaseModel):
    website: str = Field(
        description="URL of the job listing. May or may not be unique to this job."
    )
    tier: str = Field(
        description="Categorize into: Associate (general), Junior Associate, Mid-Level Associate, Senior Associate, Partner, Other (catch all if unsure)."
    )
    department: str = Field(
        description="Department, only if explicitly stated.",
    )
    title: str = Field(
        description="Job title. Remove location but keep everything else."
    )
    description: str = Field(description="Job description.")
    locations: list[str] = Field(
        description="Offices/cities/locations where this job is being offered."
    )
    salary_range: str = Field(
        description="Pay range specifically stated in the listing."
    )
    experience: str = Field(
        description="Required experience specifically stated in the listing."
    )
    full: str = Field(
        description="Complete information related to this job including HTML content for links."
    )


def get_fetch_schema(fetch_id: FetchId) -> Union[Dict[str, Any], Type[BaseModel]]:
    fetch_schemas: Dict[str, Union[Dict[str, Any], Type[BaseModel]]] = {
        "contact": ContactSchema,
        "job_posting": JobPostingSchema,
        "contract": GovernmentContractSchema,
        "manufacturing_commerce": ManufacturingCommerceSchema,
        "forum": ForumSchema,
        "attorney": AttorneySchema,
        "attorney_job_listing": AttorneyJobPostingSchema,
    }

    if fetch_id not in fetch_schemas:
        raise ValueError(f"Invalid fetch_id: {fetch_id}")

    return fetch_schemas[fetch_id]



from bananalyzer.__version import VERSION
from bananalyzer.data.examples import get_all_examples
from bananalyzer.data.schemas import Example, GoalType
from bananalyzer.runner.agent_runner import AgentRunner
from bananalyzer.runner.evals import check_match

__all__ = [
    "AgentRunner",
    "GoalType",
    "Example",
    "get_all_examples",
    "check_match",
    "VERSION",
]



import asyncio

from playwright.async_api import Page

from bananalyzer.data.schemas import Example
from bananalyzer.runner.agent_runner import AgentResult, AgentRunner


class NullAgentRunner(AgentRunner):
    """
    A test agent class that just returns an empty string
    """

    RANDOM_FAILURE_RATE = 0  # You can change this to 0.5 to see random failures

    async def run(
        self,
        page: Page,
        example: Example,
    ) -> AgentResult:
        print(f"Testing {example.get_static_url()}")
        await page.goto(example.get_static_url())
        await asyncio.sleep(0.2)

        print(f"Done testing {example.get_static_url()}")

        # Ensure page is correct
        if example.evals[0].type == "end_url_match" and isinstance(
            example.evals[0].expected, str
        ):
            await page.goto(example.evals[0].expected)
            return example.evals[0].expected

        if example.evals[0].expected is not None:
            return example.evals[0].expected  # type: ignore[return-value]
        else:
            return example.evals[0].options  # type: ignore[return-value]



import os
from typing import Any, Callable

import pytest
import xdist
from _pytest.config import Config
from _pytest.main import Session
from _pytest.nodes import Node
from _pytest.python import Function
from _pytest.terminal import TerminalReporter
from tabulate import tabulate

RecordProperty = Callable[[str, object], None]
is_github_actions = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.hookimpl(tryfirst=True)
def pytest_configure(config: Config) -> None:
    plugin = BananalyzerPytestPlugin()
    config.pluginmanager.register(plugin)


class BananalyzerPytestPlugin:
    """
    This plugin is responsible for collecting and marking properties of each test
    based on the markers that are applied to it. It also prints the results of each
    test to the terminal after all tests have been executed.

    Since we are using xdist, load this plugin through a conftest.py file so that
    it can be properly registered on each worker node.

    Dynamically imported in bananalyzer/runner/runner.py
    """

    MARKER_PREFIX = "bananalyzer_"

    def __init__(self) -> None:
        self.marks: dict[str, dict[str, str]] = {}
        self.record_property: RecordProperty = lambda x, y: None

    @pytest.hookimpl(trylast=True)
    def pytest_terminal_summary(
        self, terminalreporter: TerminalReporter, *args: Any, **kwargs: Any
    ) -> None:
        """
        This hook is called after all tests have been executed. We use it to print
        the results of each test. When using xdist this hook is only called on the
        master node.
        """

        terminalreporter.section("Bananalyzer Results")

        results: dict[str, dict[str, dict[bool, int]]] = {}
        for test_result in (
            terminalreporter.stats.get("passed", [])
            + terminalreporter.stats.get("failed", [])
            + terminalreporter.stats.get("error", [])
        ):
            for key, value in self.marks.get(test_result.nodeid, {}).items():
                result_property = results.setdefault(key, {})
                result_property_value = result_property.setdefault(value, {})
                result_property_value[test_result.passed] = (
                    result_property_value.get(test_result.passed, 0) + 1
                )

        total_passed = len(terminalreporter.stats.get("passed", []))
        total_failed = len(terminalreporter.stats.get("failed", [])) + len(
            terminalreporter.stats.get("error", [])
        )
        total_tests = total_passed + total_failed

        for k, v in results.items():
            terminalreporter.write_line(f"{k.capitalize()} Results:")
            self.print_field_data(terminalreporter, v)

        table_data = {
            "Total Tests": total_tests,
            "Tests Passed": total_passed,
            "Tests Failed": total_failed,
            "Percent Passed": f"{total_passed / total_tests * 100:.2f}%",
        }

        terminalreporter.write_line("Summary:")
        terminalreporter.write_line(
            tabulate(
                table_data.items(),
                headers=["Metric", "Value"],
                tablefmt="pipe" if is_github_actions else "psql",
            )
        )

    @pytest.fixture(autouse=True)
    def add_user_properties(self, record_property: RecordProperty) -> None:
        self.record_property = record_property

    # noinspection PyBroadException
    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_call(self, item: Function) -> None:
        """
        This hook is called prior to test execution. We use it to record the marks
        associated with each test.
        """

        def record(k: str, value: str) -> None:
            self.marks.setdefault(item.nodeid, {})[k] = value
            self.record_property(k, value)

        for key, accessor in {
            "class": lambda *_: item.parent.name,  # type: ignore
            "field": lambda *_: item.callspec.params.get("key"),
        }.items():
            try:
                record(key, accessor())  # type: ignore
            except Exception:
                pass

        for mark in [
            mark
            for mark in item.iter_markers()
            if mark.name.startswith(self.MARKER_PREFIX) and len(mark.args) == 1
        ]:
            record(mark.name[len(self.MARKER_PREFIX) :], mark.args[0])

    @pytest.hookimpl(tryfirst=True)
    def pytest_sessionstart(self, session: Session) -> None:
        """
        If using xdist, we need to emit the marks from each worker to the master. This
        hook can be called on both the worker and master nodes.
        """
        if xdist.is_xdist_worker(session):
            session.config.workeroutput["marks"] = self.marks  # type: ignore

    @pytest.hookimpl(trylast=True)
    def pytest_testnodedown(self, node: Node, *_: Any) -> None:
        """
        If using xdist, we need to collect the marks from each worker. This hook is
        called on the master node only.
        """
        if "marks" in node.workeroutput:  # type: ignore
            node_stats = node.workeroutput["marks"]  # type: ignore
            self.marks.update(node_stats)

    @staticmethod
    def print_field_data(
        terminalreporter: TerminalReporter, results: dict[str, dict[bool, int]]
    ) -> None:
        headers = "Group", "Passed", "Failed", "Correct %", "Perfect"

        table_data = []
        total_passed = 0
        total_failed = 0
        perfect_count = 0

        for field, values in results.items():
            passed = values.get(True, 0)
            failed = values.get(False, 0)
            total_passed += passed
            total_failed += failed
            is_perfect = failed == 0
            perfect_count += int(is_perfect)
            percentage = (
                f"{passed / (passed + failed) * 100:.2f}%"
                if (passed + failed) > 0
                else "N/A"
            )
            perfect = is_perfect and "✅" or "❌"
            table_data.append([field, passed, failed, percentage, perfect])

        table_data.sort(key=lambda row: row[3], reverse=True)  # type: ignore

        # Calculate the percentage for the total row
        total_percentage = (
            f"{total_passed / (total_passed + total_failed) * 100:.2f}%"
            if (total_passed + total_failed) > 0
            else "N/A"
        )
        perfect_percentage = (
            f"{perfect_count / len(results) * 100:.2f}%" if results else "N/A"
        )

        # Add a totals row
        table_data.append(["-" * len(header) for header in headers])
        table_data.append(
            ["Total", total_passed, total_failed, total_percentage, perfect_percentage]
        )

        # Create a table using tabulate
        table = tabulate(
            table_data,
            headers=headers,
            tablefmt="pipe" if is_github_actions else "psql",
        )

        # Print the table
        terminalreporter.write_line(table)
        terminalreporter.write_line("\n")
